"""advent of code runner for creating CLI for solution for easy execution"""
import click


def advent_runner(func):
    """
    Advent of code runner decorator which decorate out a function for a simple CLI
    tool creation. Function which is decorated using a advent of code runner should
    return out a Initializer class so it can create out a CLI from Initializer class
    object
    """

    def create_command():
        initializer_defined = func()

        @click.group(help="CLI tool created from a provided Initializer Object")
        def main():
            pass

        @main.command(
            "run", help="run a certain function by name present in Initializer"
        )
        @click.argument("name")  # pylint:disable=unused-variable
        def run_function(name):
            initializer_defined.run(name)

        @main.command("all", help="run all function defined in a Initializer class")
        def run_all_function():  # pylint:disable=unused-variable
            initializer_defined.run_all()

        @main.command("list", help="list out all of the available function")
        def list_functions():  # pylint:disable=unused-variable
            initializer_defined.list_functions()

        return main()

    return create_command



"""module used for defining puzzle decorator for submiting or solving problem"""
from typing import Callable, List, Optional, TypeVar

from pydantic import BaseModel

from .cache_file import cache_file_data
from .config_file import get_all_session
from .server_action import submit_output

T = TypeVar("T")  # pylint:disable=invalid-name


class _Puzzle(BaseModel):
    """Puzzle class for handling out a puzzle decorator"""

    function: Callable[[str], T]
    year: int
    day: int
    part: int
    session: List[str] = get_all_session()
    operation_type: str
    input_file: Optional[str] = None

    def __repr__(self):
        """return repr value of class which is set to function __name__"""
        return "{}".format(self.function.__name__)

    def getfunction(self):
        """get out function from a class"""
        return self.function

    def __call__(self):
        """Caller for _Puzzle class"""
        for session_list in self.session:
            if self.input_file is None:
                input_data = cache_file_data(self.year, self.day, session_list)
            else:
                with open(self.input_file) as opened_file:
                    input_data = opened_file.read()
            answer = self.function(input_data)
            if answer is not None:
                if self.operation_type == "submit":
                    message = submit_output(
                        self.year, self.day, self.part, session_list, answer
                    )
                    if message.contains("Congratulation"):
                        print(
                            "{}:{}-{}-{}: {} {} {}".format(
                                session_list,
                                self.year,
                                self.day,
                                self.part,
                                answer,
                                u"\u2713",
                                message,
                            )
                        )
                    else:
                        print(
                            "{}:{}-{}-{}: {} {} {}".format(
                                session_list,
                                self.year,
                                self.day,
                                self.part,
                                answer,
                                u"\u274C",
                                message,
                            )
                        )
                elif self.operation_type == "solve":
                    print(
                        "{}:{}-{}-{}: {}".format(
                            session_list, self.year, self.day, self.part, answer
                        )
                    )


def submit(
    year: int,
    day: int,
    part: int,
    session_list: Optional[List[str]] = None,
    input_file: Optional[str] = None,
):
    """
    Puzzle decorator used to submit a solution to advent_of_code server and provide
    result. If input_file is not present then it tries to download file and cache it
    for submiting solution else it require to be provided with input_file path which
    input it can use out
    """

    def _action(function):
        operation_type = "submit"
        if not session_list:
            session = get_all_session()
        else:
            session = session_list
        return _Puzzle(
            function=function,
            operation_type=operation_type,
            year=year,
            day=day,
            part=part,
            session=session,
            input_file=input_file,
        )

    return _action


def solve(
    year: int,
    day: int,
    part: int,
    session_list: Optional[List[str]] = None,
    input_file: Optional[str] = None,
):
    """
    Puzzle decorator used to only solve a solution & print output value. It doesn't
    submit output to advent of code server to validate out whether an answer is correct
    or not. By default it downloads out input file from advent-of-code server. Puzzle
    can also be solved by using custom file location with help of input_file parameter
    while using input_file year, day, part are required even if they are not used for
    only reference purpose of printing output.
    """

    def _action(function):
        operation_type = "solve"
        if not session_list:
            session = get_all_session()
        else:
            session = session_list
        return _Puzzle(
            function=function,
            operation_type=operation_type,
            year=year,
            day=day,
            part=part,
            session=session,
            input_file=input_file,
        )

    return _action



"""utility module used to find latest year and day for advent of code problems"""
import datetime

from dateutil.tz import gettz

EASTERN = gettz("America/New_York")


def get_current_year() -> int:
    """
    Get latest year of advent of code return old year if month is not December
    otherwise return current year
    """
    date_time = datetime.datetime.now(tz=EASTERN)
    if date_time.month < 12:
        return date_time.year - 1
    return date_time.year


def get_day() -> int:
    """Get latest day or set a latest day as 25 if this is not month of December"""
    date_time = datetime.datetime.now(tz=EASTERN)
    if date_time.month == 12:
        return min(date_time.day, 25)
    return 25



"""Module to perform config file related operation"""
import json
import os
from pathlib import Path
from typing import List

import appdirs


class Data:
    """Class to store data about a config file of advent-of-code"""

    def __init__(self, **kwargs: str):
        """Initialize Data class"""
        self.session_list = kwargs

    def to_json(self) -> str:
        """Convert Data to JSON type"""
        return json.dumps(self.__dict__, indent=4, sort_keys=True)

    @classmethod
    def from_json(cls, json_str):
        """Create Data type from JSON"""
        json_dict = json.loads(json_str)
        json_data = json_dict.get("session_list")
        return cls(**json_data)

    def add_session(self, **kwargs: str):
        """Add new session or update session in a config file Data type."""
        self.session_list.update(kwargs)

    def delete_session(self, session: str):
        """Delete Session from config files"""
        try:
            self.session_list.pop(session)
        except KeyError:
            raise Exception("Value with name {} not found over config".format(session))


def _read_json_file(config_file: str) -> Data:
    """Read JSON file to return Data type"""
    if not config_file.exists():
        with open(config_file, "a+") as opened_file:
            val = Data()
            opened_file.write(val.to_json())
    with open(config_file) as opened_file:
        val = opened_file.read()
        data = Data.from_json(val)
    return data


def _config_file_data() -> str:
    """Get Config file location"""
    config_location = appdirs.user_config_dir()
    config_file = os.path.join(config_location, "aoc-config.json")
    config_file = Path(config_file)
    return config_file


def add_to_json(**kwargs: str):
    """Add new session to json config file"""
    config_file = _config_file_data()
    data = _read_json_file(config_file)
    data.add_session(**kwargs)
    with open(config_file, "w") as opened_file:
        opened_file.write(data.to_json())


def delete_from_json(session: str):
    """Delete session from JSON config file"""
    config_file = _config_file_data()
    data = _read_json_file(config_file)
    data.delete_session(session)
    with open(config_file, "w") as opened_file:
        opened_file.write(data.to_json())


def list_from_json():
    """List all session from a JSON file along with value"""
    config_file = _config_file_data()
    data = _read_json_file(config_file)
    json_dict = json.loads(data.to_json())
    json_data = json_dict.get("session_list")
    for key, val in json_data.items():
        print("{} = {}".format(key, val))


def get_session_value(session_name: str) -> str:
    """Return session name value from JSON config file"""
    config_file = _config_file_data()
    data = _read_json_file(config_file)
    json_dict = json.loads(data.to_json())
    json_data = json_dict.get("session_list")
    session_value = json_data.get(session_name)
    if session_value is not None:
        return session_value
    raise Exception("{} key is not present in config file".format(session_name))


def get_all_session() -> List[str]:
    """Return all session name"""
    config_file = _config_file_data()
    data = _read_json_file(config_file)
    json_dict = json.loads(data.to_json())
    json_data = json_dict.get("session_list")
    session_list = list(json_data.keys())
    return session_list



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['advent_of_code_py']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'click>=7.0,<8.0',
 'pydantic>=1.4,<2.0',
 'python-dateutil>=2.8.0,<3.0.0',
 'requests>=2.22.0,<3.0.0']

entry_points = \
{'console_scripts': ['advent-of-code-py = advent_of_code_py.cli:main']}

setup_kwargs = {
    'name': 'advent-of-code-py',
    'version': '0.2.0',
    'description': 'Advent of Code helper CLI and library',
    'long_description': '# Advent-of-code-py\n[Advent of Code][advent_of_code_link] helper CLI and library for python projects.\n\n**Note:**\nCurrently it is still in beta stage which may have lots of bug please report out bug if you find any\n\n**Status & Info:**\n\n| Travis Build Status | Code style | License | Project Version |\n| :---: | :---: | :---: | :---: |\n| [![Travis Build Status][build_badge]][build_link] | [![Code style][black_badge]][black_link] | [![License: MIT][license_badge]][license_link] | [![PyPI][project_badge]][project_link] |\n\n## Usage\n\n### Installation\nTo install out advent-of-code-py run out following command which install out advent-of-code-py cli and advent_of_code_py library.\n```bash\npip install advent-of-code-py\n```\n\n__OR__\n\n```bash\npoetry add advent-of-code-py\n```\n\n### Usage\nInitially for advent-of-code-py to work out it need session value or session ID which you can obtain out by viewing out cookie while visiting advent of code server.\nAfter collecting session cookie value you need to add those value in config using advent-of-code-py CLI\n```bash\nadvent-of-code-py config add <session-name> <session-value>\n```\n\nNow you can import out library by using\n```python\nimport advent_of_code_py\n```\n\nAfter importing a library you can now use out either two decorator solve or submit decorator for a function of puzzle\n\nFor example:-\n```python\n@advent_of_code_py.submit(2018,3,1,session_list="<session-name>")\ndef puzzle_2018_3_1(data=None):\n    # do some calculation with data and return final output\n    return final_output\n```\n\nNow after decorating out function now you can call out function\n```python\npuzzle_2018_3_1()\n```\nAfter calling out function `final_output` value will be submitted out to Advent of Code server for 2018 year day 3\nproblem then returns out whether submitted answer was correct or not. If session value is not provided then\nsolution will be submitted to all session value set out.\n\nYou can also use out advent-of-code-py builtin Initializer and runner to create out appropriate CLI\ntool for problem so problem can be run easily\nTo set advent-of-code-py puzzle as CLI\n```python\n@advent_of_code_py.advent_runner()\ndef main_cli():\n    initializer = advent_of_code_py.Initializer()\n    initializer.add("<function_alias>"="<function>")\n    # for example to run above function you can write out\n    initializer.add(p_3_1=puzzle_2018_3_1)\n    # add other function\n    return initializer\n```\nNow you can set out main_cli as entry points and it will create out CLI with appropriate name and function which was added.\nSo for example to run out function puzzle_2018_3_1() you have to run out command as `entry-point-name run p_3_1` which\nwill run appropriate function as well as submit as desired.\n\n[advent_of_code_link]: https://adventofcode.com\n\n[build_badge]: https://img.shields.io/travis/com/iamsauravsharma/advent-of-code-py.svg?logo=travis\n[build_link]: https://travis-ci.com/iamsauravsharma/advent-of-code-py\n\n[black_badge]: https://img.shields.io/badge/code%20style-black-000000.svg\n[black_link]: https://github.com/ambv/black\n\n[license_badge]: https://img.shields.io/github/license/iamsauravsharma/advent-of-code-py.svg\n[license_link]: LICENSE\n\n[project_badge]: https://img.shields.io/pypi/v/advent-of-code-py?color=blue&logo=python\n[project_link]: https://pypi.org/project/advent-of-code-py',
    'author': 'Saurav Sharma',
    'author_email': 'appdroiddeveloper@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/iamsauravsharma/advent-of-code-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
